/*
 *
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/connection.h"
#include <grpcpp/create_channel.h>                // for CreateChannel
#include <grpcpp/impl/codegen/client_context.h>   // for ClientContext
#include <grpcpp/impl/codegen/status.h>           // for Status
#include <grpcpp/security/credentials.h>          // for SslCredentials
#include <iostream>                               // for operator<<, basic_...
#include <stdexcept>                              // for runtime_error
#include <string>                                 // for operator+, char_tr...
#include "caosdb/authentication.h"                // for Authenticator
#include "caosdb/exceptions.h"                    // for AuthenticationError
#include "caosdb/info/v1alpha1/main.grpc.pb.h"    // for GeneralInfoService
#include "caosdb/info/v1alpha1/main.pb.h"         // for GetVersionInfoResp...
#include "caosdb/utils.h"                         // for load_string_file
#include "caosdb/info.h"                          // for VersionInfo
#include "grpcpp/impl/codegen/status_code_enum.h" // for StatusCode, UNAUTH...
namespace grpc {
class Channel;
} // namespace grpc

namespace caosdb::connection {
using caosdb::authentication::Authenticator;
using caosdb::exceptions::AuthenticationError;
using caosdb::exceptions::ConnectionError;
using caosdb::info::VersionInfo;
using caosdb::info::v1alpha1::GeneralInfoService;
using caosdb::info::v1alpha1::GetVersionInfoRequest;
using caosdb::info::v1alpha1::GetVersionInfoResponse;
using caosdb::utils::load_string_file;
using grpc::InsecureChannelCredentials;
using grpc::SslCredentials;
using grpc::SslCredentialsOptions;

PemFileCACertProvider::PemFileCACertProvider(const std::string &path) {
  this->cacert = load_string_file(path);
}

auto PemFileCACertProvider::getCACertPem() const -> std::string {
  return this->cacert;
}

PemCACertProvider::PemCACertProvider(const std::string &cacert) {
  this->cacert = cacert;
}

auto PemCACertProvider::getCACertPem() const -> std::string {
  return this->cacert;
}

CaosDBConnectionConfig::CaosDBConnectionConfig(const std::string &host,
                                               int port) {
  this->host = host;
  this->port = port;
}

auto CaosDBConnectionConfig::getHost() const -> std::string {
  return this->host;
}

auto CaosDBConnectionConfig::getPort() const -> int { return this->port; }

auto operator<<(std::ostream &out, const CaosDBConnectionConfig &config)
  -> std::ostream & {
  out << config.toString();
  return out;
}

InsecureCaosDBConnectionConfig::InsecureCaosDBConnectionConfig(
  const std::string &host, int port)
  : CaosDBConnectionConfig(host, port) {
  this->credentials = InsecureChannelCredentials();
}

InsecureCaosDBConnectionConfig::InsecureCaosDBConnectionConfig(
  const std::string &host, int port,
  const std::shared_ptr<Authenticator> &authenticator)
  : CaosDBConnectionConfig(host, port) {
  this->credentials = grpc::CompositeChannelCredentials(
    InsecureChannelCredentials(), authenticator->getCallCredentials());
}

auto InsecureCaosDBConnectionConfig::getChannelCredentials() const
  -> std::shared_ptr<ChannelCredentials> {
  return this->credentials;
}

auto InsecureCaosDBConnectionConfig::toString() const -> std::string {
  return "InsecureCaosDBConnectionConfig(" + this->getHost() + "," +
         std::to_string(this->getPort()) + ")";
}

SslCaosDBConnectionConfig::SslCaosDBConnectionConfig(
  const std::string &host, int port,
  const std::shared_ptr<CACertificateProvider> &cacert)
  : CaosDBConnectionConfig(host, port) {
  SslCredentialsOptions options;
  options.pem_root_certs = cacert->getCACertPem();
  std::cout << "HERE" << options.pem_root_certs;
  this->credentials = SslCredentials(options);
}

SslCaosDBConnectionConfig::SslCaosDBConnectionConfig(
  const std::string &host, int port,
  const std::shared_ptr<CACertificateProvider> &cacert,
  const std::shared_ptr<Authenticator> &authenticator)
  : CaosDBConnectionConfig(host, port) {

  SslCredentialsOptions options;
  options.pem_root_certs = cacert->getCACertPem();
  this->credentials = grpc::CompositeChannelCredentials(
    SslCredentials(options), authenticator->getCallCredentials());
}

auto SslCaosDBConnectionConfig::getChannelCredentials() const
  -> std::shared_ptr<ChannelCredentials> {
  return this->credentials;
}

auto SslCaosDBConnectionConfig::toString() const -> std::string {
  return "SslCaosDBConnectionConfig(" + this->getHost() + "," +
         std::to_string(this->getPort()) + ")";
}

CaosDBConnection::CaosDBConnection(
  const std::shared_ptr<CaosDBConnectionConfig> &config) {
  this->config = config;
  const std::string target =
    this->config->getHost() + ":" + std::to_string(this->config->getPort());
  const std::shared_ptr<grpc::Channel> &channel =
    grpc::CreateChannel(target, this->config->getChannelCredentials());
  this->stub_ = GeneralInfoService::NewStub(channel);
}

auto operator<<(std::ostream &out, const CaosDBConnection &connection)
  -> std::ostream & {
  out << "CaosDBConnection(" << *(connection.config) << ")";
  return out;
}

[[nodiscard]] auto CaosDBConnection::getVersionInfo() const
  -> std::unique_ptr<VersionInfo> {
  GetVersionInfoRequest request;
  GetVersionInfoResponse response;
  grpc::ClientContext context;
  const grpc::Status status =
    stub_->GetVersionInfo(&context, request, &response);

  if (!status.ok()) {
    switch (status.error_code()) {
    case grpc::StatusCode::UNAUTHENTICATED:
      throw AuthenticationError(status.error_message());
    case grpc::StatusCode::UNAVAILABLE:
      throw ConnectionError(status.error_message());
    default:
      std::cout << status.error_code() << "\n";
      throw std::runtime_error(status.error_message());
    }
  }
  return std::make_unique<VersionInfo>(response.release_version_info());
}

} // namespace caosdb::connection
